from tira.third_party_integrations import extract_to_be_executed_notebook_from_command_or_none, extract_previous_stages_from_notebook, parse_ast_extract_assignment, parse_extraction_of_tira_approach, extract_previous_stages_from_docker_image, parse_extraction_of_tira_approach_bash
from pathlib import Path
import unittest
from subprocess import check_output

TEST_DIR = Path(__file__).parent.resolve()

class JupyterNotebookPipelineConstructionTest(unittest.TestCase):
    def test_no_notebook_is_extracted_for_none_command(self):
        command = None
        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertIsNone(actual)

    def test_no_notebook_is_extracted_for_empty_command(self):
        command = ''
        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertIsNone(actual)

    def test_notebook_is_extracted_for_pyterrier_command(self):
        expected = '/workspace/notebook.ipynb'
        command = '/workspace/run-pyterrier-notebook.py --input $inputDataset --output $outputDir --notebook /workspace/notebook.ipynb'

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_notebook_is_extracted_for_command_in_between(self):
        expected = '/notebook.ipynb'
        command = '/workspace/run-notebook.py --notebook /notebook.ipynb --input $inputDataset --output $outputDir'

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_py_file_is_extracted_for_command_in_between(self):
        expected = 'app.py'
        command = 'python3 app.py'

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_no_previous_stages_are_extracted_from_notebook_without_previous_stages(self):
        notebook = TEST_DIR / 'resources' / 'pyterrier-notebook-without-previous-stages.ipynb'
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_no_previous_stages_are_extracted_from_python_file_without_previous_stages(self):
        notebook = TEST_DIR / 'resources' / 'pyterrier-notebook-without-previous-stages.py'
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage(self):
        notebook = TEST_DIR / 'resources' / 'retrieve-with-pyterrier-index.ipynb'
        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_python_file(self):
        notebook = TEST_DIR / 'resources' / 'retrieve-with-pyterrier-index.py'
        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_parsing_of_ast_assignment_none(self):
        python_line = None
        k, v =  parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_empty_string(self):
        python_line = None
        k, v =  parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_method_call(self):
        python_line = "index = tira.pt.index('ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)', 'longeval-tiny-train-20240315-training')"
        k, v =  parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_to_constant(self):
        python_line = "index = 3"
        expected_k, expeced_v = 'index', 3
        k, v =  parse_ast_extract_assignment(python_line)

        self.assertEqual(expected_k, k)
        self.assertEqual(expeced_v, v)

    def test_extraction_of_approach_is_failsave_01(self):
        python_line = "index = None"
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_02(self):
        python_line = ""
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_03(self):
        python_line = None
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_04(self):
        python_line = "#index = tira.pt.index('tmp', 'longeval-tiny-train-20240315-training') # some comment"
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_01(self):
        python_line = "index = tira.pt.index('ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)', 'longeval-tiny-train-20240315-training')"
        expected = 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)'
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_02(self):
        python_line = "index = tira.pt.index('tmp', 'longeval-tiny-train-20240315-training') # some comment"
        expected = 'tmp'
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_03(self):
        python_line = 'index = tira.pt.index("tmp\'\\"a","longeval-tiny-train-20240315-training") # some comment'
        expected = 'tmp\'"a'
        actual =  parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_integration_against_docker_image_01(self):
        image = 'mam10eks/bash-with-notebooks:latest'
        command = 'bash /workspace/run-notebook.sh --notebook /pyterrier-notebook-without-previous-stages.ipynb --input $inputDataset --output $outputDir'

        expected = []
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)

    def test_integration_against_docker_image_02(self):
        image = 'mam10eks/bash-with-notebooks:latest'
        command = 'bash /workspace/run-notebook.sh  --input $inputDataset --output $outputDir --notebook /pyterrier-notebook-without-previous-stages.ipynb'

        expected = []
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)


    def test_integration_against_docker_image_03(self):
        image = 'mam10eks/bash-with-notebooks:latest'
        command = 'bash /workspace/run-notebook.sh  --input $inputDataset --output $outputDir --notebook /retrieve-with-pyterrier-index.ipynb'

        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_01(self):
        bash_line = ""
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_02(self):
        bash_line = None
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_03(self):
        bash_line = 'tira-cli --help'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_01(self):
        bash_line = "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach tmp) # some comment"
        expected = 'tmp'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_02(self):
        bash_line = "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)') # some comment"
        expected = 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_03(self):
        bash_line = "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach \"tmp\") # some comment"
        expected = 'tmp'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_04(self):
        bash_line = "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach \"ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)\") # some comment"
        expected = 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_05(self):
        bash_line = "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach 'tmp') # some comment"
        expected = 'tmp'
        actual =  parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_bash_file(self):
        notebook = TEST_DIR / 'resources' / 'retrieve-with-pyterrier-bash.sh'
        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_no_previous_stage_for_bash_file(self):
        notebook = TEST_DIR / 'resources' / 'retrieve-no-previous-stages-bash.sh'
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_01(self):
        image = 'dockerfile_bash_script_absolute'
        check_output(['docker', 'build', '-t', image, '-f', f'tests/resources/{image}', '.'])

        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']
        actual = extract_previous_stages_from_docker_image(image, '/usr/bin/retrieve-with-pyterrier-bash.sh')

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_02(self):
        image = 'dockerfile_bash_script_absolute'
        check_output(['docker', 'build', '-t', image, '-f', f'tests/resources/{image}', '.'])

        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']
        actual = extract_previous_stages_from_docker_image(image)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_03(self):
        image = 'dockerfile_bash_script_absolute'
        check_output(['docker', 'build', '-t', image, '-f', f'tests/resources/{image}', '.'])

        expected = []
        actual = extract_previous_stages_from_docker_image(image, '/etc/hostname')

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_04(self):
        image = 'jupyter_script_relative'
        check_output(['docker', 'build', '-t', image, '-f', f'tests/resources/{image}', '.'])

        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']
        actual = extract_previous_stages_from_docker_image(image)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_05(self):
        image = 'jupyter_script_relative'
        check_output(['docker', 'build', '-t', image, '-f', f'tests/resources/{image}', '.'])

        expected = ['ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)']
        actual = extract_previous_stages_from_docker_image(image, 'python3 /usr/bin/retrieve-with-pyterrier-index.py')

        self.assertEqual(expected, actual)


class PandasIntegration():
    def __init__(self, tira_client):
        self.tira_client = tira_client

    def from_retriever_submission(self, approach, dataset, previous_stage=None, datasets=None):
        import pandas as pd
        from tira.ir_datasets_util import translate_irds_id_to_tirex
        task, team, software = approach.split('/')

        if dataset and datasets:
            raise ValueError(f'You can not pass both, dataset and datasets. Got dataset = {dataset} and datasets= {datasets}')

        if not datasets:
            datasets = [dataset]

        df_ret = []
        for dataset in datasets:
            ret, run_id = self.tira_client.download_run(task, translate_irds_id_to_tirex(dataset), software, team, previous_stage, return_metadata=True)
            ret['qid'] = ret['query'].astype(str)
            ret['docno'] = ret['docid'].astype(str)
            del ret['query']
            del ret['docid']

            ret['tira_task'] = task
            ret['tira_dataset'] = dataset
            ret['tira_first_stage_run_id'] = run_id
            df_ret += [ret]

        return pd.concat(df_ret)

    def __matching_files(self, approach, dataset, file_selection):
        from glob import glob
        from tira.ir_datasets_util import translate_irds_id_to_tirex
        ret = set()
        
        if type(file_selection) is str:
            file_selection = [file_selection]
        
        for glob_entry in file_selection:
            glob_entry = self.tira_client.get_run_output(approach, (translate_irds_id_to_tirex(dataset))) + glob_entry
            for i in glob(glob_entry): ret.add(i)
        
        return sorted(list(ret))

    def transform_queries(self, approach, dataset, file_selection=('/*.jsonl', '/*.jsonl.gz')):
        import pandas as pd
        matching_files = self.__matching_files(approach, dataset, file_selection)
        
        if len(matching_files) == 0:
            raise ValueError(f'Could not find a matching query output. Found: {matching_files}. Please specify the file_selection to resolve this.')

        ret = pd.read_json(matching_files[0], lines=True, dtype={'qid': str, 'query': str, 'query_id': str})
        if 'qid' not in ret and 'query_id' in ret:
            ret['qid'] = ret['query_id']
            del ret['query_id']

        return ret

    def transform_documents(self, approach, dataset, file_selection=('/*.jsonl', '/*.jsonl.gz')):
        import pandas as pd
        matching_files = self.__matching_files(approach, dataset, file_selection)
        if len(matching_files) == 0:
            raise ValueError('Could not find a matching document output. Used file_selection: ' + str(file_selection) + '. Please specify the file_selection to resolve this.')

        ret = pd.read_json(matching_files[0], lines=True, dtype={'docno': str, 'doc_id': str})

        if 'doc_id' in ret.columns and 'docno' not in ret.columns:
            ret['docno'] = ret['doc_id']
            del ret['doc_id']
        return ret



from abc import ABC
from typing import TYPE_CHECKING, Union, overload

if TYPE_CHECKING:
    import io
    from typing import Any, Dict, Optional

    import pandas as pd

# .. todo:: this file needs further documentation


class TiraClient(ABC):
    def all_datasets(self) -> "pd.DataFrame":
        pass

    def all_softwares(self) -> "pd.DataFrame":
        pass

    def print_overview_of_all_software(self) -> None:
        pass

    def all_evaluators(self) -> "pd.DataFrame":
        pass

    def all_evaluated_appraoches(self) -> "pd.DataFrame":
        pass

    def docker_software() -> "Any":
        # .. todo:: typehint
        pass

    def run_was_already_executed_on_dataset(self, approach, dataset) -> bool:
        # .. todo:: typehint
        pass

    def get_run_output(self, approach, dataset, allow_without_evaluation=False) -> str:
        # .. todo:: typehint
        pass

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None) -> "Optional[Dict[str, str]]":
        # .. todo:: typehint
        pass

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "tuple[pd.DataFrame, str]":
        # .. todo:: typehint
        ...

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "pd.DataFrame":
        # .. todo:: typehint
        ...

    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "Union[pd.DataFrame, tuple[pd.DataFrame, str]]":
        # .. todo:: typehint
        pass

    def create_new_upload(self, task_id: str, vm_id: str) -> "Optional[str]":
        """
        Creates a new upload and returns the newly created id. Returns None on failure.
        """
        pass

    def submit_run(self, task_id: str, vm_id: str, dataset_id: str, upload_id: str, run: "io.IOBase") -> bool:
        """
        Submits the runfile located at `run` for the given task and vm for the given upload id. Returns true on success,
        false  otherwise.
        """
        pass



from tira.io_utils import all_environment_variables_for_github_action_or_fail
import unittest


class TestExportSubmissionEnvironment(unittest.TestCase):
    def test_empty_parameters_fails(self):
        with self.assertRaises(ValueError):
            all_environment_variables_for_github_action_or_fail([])

    def test_all_required_parameters_are_there_docker_image(self):
        actual = all_environment_variables_for_github_action_or_fail(['GITHUB_SHA=sha', 'TIRA_VM_ID=vm', 'TIRA_TASK_ID=task', 'TIRA_DOCKER_REGISTRY_TOKEN=docker-token', 'TIRA_DOCKER_REGISTRY_USER=docker-user', 'TIRA_CLIENT_TOKEN=tira-token', 'TIRA_CLIENT_USER=tira-user', 'TIRA_CODE_REPOSITORY_ID=repo-id', 'TIRA_DOCKER_FILE=Dockerfile', 'TIRA_DOCKER_PATH=dockerpath', 'TIRA_COMMAND=cmd'])

        self.assertTrue('GITHUB_SHA=sha' in actual)
        self.assertTrue('IMAGE_TAG=registry.webis.de/code-research/tira/tira-user-vm/submission-dockerpath:sha' in actual)

    def test_all_required_parameters_are_there_jupyter_notebook(self):
        actual = all_environment_variables_for_github_action_or_fail(['GITHUB_SHA=sha', 'TIRA_VM_ID=vm', 'TIRA_TASK_ID=task', 'TIRA_DOCKER_REGISTRY_TOKEN=docker-token', 'TIRA_DOCKER_REGISTRY_USER=docker-user', 'TIRA_CLIENT_TOKEN=tira-token', 'TIRA_CLIENT_USER=tira-user', 'TIRA_CODE_REPOSITORY_ID=repo-id', 'TIRA_DOCKER_FILE=Dockerfile', 'TIRA_DOCKER_PATH=dockerpath', 'TIRA_COMMAND=cmd', 'TIRA_JUPYTER_NOTEBOOK=123'])

        self.assertTrue('GITHUB_SHA=sha' in actual)
        self.assertTrue('IMAGE_TAG=registry.webis.de/code-research/tira/tira-user-vm/submission:sha' in actual)




import json
import pandas as pd
import os
from glob import glob
from packaging import version
from tira.pyterrier_integration import PyTerrierIntegration
from tira.pandas_integration import PandasIntegration
from tira.local_execution_integration import LocalExecutionIntegration
from .tira_client import TiraClient


class Client(TiraClient):
    def __init__(self, directory='.', rest_client=None):
        self.pd = PandasIntegration(self)
        self.pt = PyTerrierIntegration(self)
        self.directory = directory + '/'
        self.tira_cache_dir = os.environ.get('TIRA_CACHE_DIR', os.path.expanduser('~') + '/.tira')
        self.rest_client = rest_client
        self.local_execution = LocalExecutionIntegration(self)

    def all_datasets(self) -> pd.DataFrame:
        ret = []
        for i in glob(self.directory + '*/training-datasets/'):
            cnt = 0
            for j in glob(i + '*'):
                cnt += len(list(open(j)))

            ret += [{'dataset': i.split('/training-datasets/')[0], 'records': cnt}]

        for i in set([j.split('/')[1] for j in glob(self.directory + '**/**/**/job-executed-on-*.txt')]):
            ret += [{'dataset': i}]

        return pd.DataFrame(ret).sort_values('dataset')

    def ___load_softwares(self):
        softwares = [json.loads(i) for i in open(self.directory + '.tira/submitted-software.jsonl')]

        return {i['TIRA_TASK_ID'] + '/' + i['TIRA_VM_ID'] + '/' + i['TIRA_SOFTWARE_NAME']: i for i in softwares}

    def all_softwares(self) -> pd.DataFrame:
        ret = []
        for software_id, software_definition in self.___load_softwares().items():
            ret += [{'approach': software_id, 'id': str(int(software_definition['TIRA_SOFTWARE_ID'].split('docker-software-')[1])), 'team': software_definition['TIRA_VM_ID'], 'image': software_definition['TIRA_IMAGE_TO_EXECUTE_IN_DOCKERHUB'], 'command': software_definition['TIRA_COMMAND_TO_EXECUTE'], 'ids_of_previous_stages': [str(int(i)) for i in software_definition['TIRA_IDS_OF_PREVIOUS_STAGES']]}]

        return pd.DataFrame(ret)

    def print_overview_of_all_software(self) -> None:
        pre_text = """# Software for [$TASK_ID](https://www.tira.io/task/$TASK_ID)

<p id="instructions">Place <code>todo-replace</code> in a directory <code>$PWD/input</code>.
Then run the software as described below to produce results in <code>$PWD/output</code>.</p>
<p>For each software three ways to run it are shown below:</p>
<ul>
<li id="description-docker"><b>Docker</b>: on the command line (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
)</li>
<li id="description-tira-cli"><b>TIRA (CLI)</b>: on the command line with <code>tira-run</code> (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
and the
<a href="https://pypi.org/project/tira/">Python TIRA package</a>
)</li>
<li id="description-tira-python"><b>TIRA (Python)</b>: in a Python script (requires
<a href="https://docs.docker.com/engine/installation/">Docker</a>
and the
<a href="https://pypi.org/project/tira/">Python TIRA package</a>
)</li>
</ul>
"""

        toc = ['## List of software']
        template_toc_team = '- [Team \u0060$TEAM\u0060](#team-$TEAM_LINK)'
        template_toc_software = '  - [Software \u0060$SOFTWARE\u0060](#software-$SOFTWARE_LINK)'

        content = []
        template_team_header = '\n## Team \u0060$TEAM\u0060\n[See generic instructions above](#instructions)'
        template_software = """### Software \u0060$SOFTWARE\u0060
- [Docker](#description-docker):
  \u0060\u0060\u0060bash
  $CMD_TIRA_DOCKER
  \u0060\u0060\u0060
- [TIRA (CLI)](#description-tira-cli):
  \u0060\u0060\u0060bash
  $CMD_TIRA_CLI
  \u0060\u0060\u0060
- [TIRA (Python)](#description-tira-python):
  \u0060\u0060\u0060python
  $CMD_TIRA_PYTHON
  \u0060\u0060\u0060
"""
        separator = '---'

        softwares = self.all_softwares().sort_values(by=['approach'])

        prev_team = ''
        for _, i in softwares.iterrows():
            execution_info = self.local_execution.run(
                identifier=i['approach'], input_dir='$PWD/input',
                output_dir='$PWD/output', dry_run=True
            )

            software_name = i["approach"].split("/")[-1]
            software_link = software_name.lower().replace(' ', '-')
            team_name = i["team"]
            team_link = team_name.lower().replace(' ', '-')

            if team_name != prev_team:
                toc.append(template_toc_team.replace('$TEAM_LINK', team_link).replace('$TEAM', team_name))

                if len(prev_team) == 0:
                    pre_text = pre_text.replace('$TASK_ID', '/'.join(i["approach"].split("/")[:-2]))
                else:
                    content.append(separator)
                content.append(template_team_header.replace('$TEAM', team_name))
                prev_team = team_name

            toc.append(template_toc_software.replace('$SOFTWARE_LINK', software_link).replace('$SOFTWARE', software_name))

            content.append(
                template_software
                .replace('$SOFTWARE', software_name)
                .replace('$CMD_TIRA_CLI', execution_info['tira-run-cli'])
                .replace('$CMD_TIRA_PYTHON', execution_info['tira-run-python'])
                .replace('$CMD_TIRA_DOCKER', execution_info['docker'])
            )

        print(pre_text)
        print('\n'.join(toc))
        print('\n'.join(content))

    def __load_evaluators(self):
        evaluators = [json.loads(i) for i in open(self.directory + '.tira/evaluators.jsonl')]
        ret = {i['TIRA_DATASET_ID']: i for i in evaluators}

        for evaluator in evaluators:
            dataset_id = evaluator['TIRA_DATASET_ID']
            current_version = version.parse(ret[dataset_id]['TIRA_EVALUATION_IMAGE_TO_EXECUTE'].split(':')[-1])
            available_version = version.parse(evaluator['TIRA_EVALUATION_IMAGE_TO_EXECUTE'].split(':')[-1])

            if available_version > current_version:
                ret[dataset_id] = evaluator

        return ret

    def __load_job_data(self, job_file):
        job = [i.split('=', 1) for i in open(job_file, 'r')]
        return {k.strip(): v.strip() for k, v in job}

    def all_evaluators(self) -> pd.DataFrame:
        ret = []
        for i in self.__load_evaluators().values():
            ret += [{'dataset': i['TIRA_DATASET_ID'], 'image': i['TIRA_EVALUATION_IMAGE_TO_EXECUTE'], 'command': i['TIRA_EVALUATION_COMMAND_TO_EXECUTE']}]

        return pd.DataFrame(ret)

    def __extract_image_and_command(self, identifier, evaluator=False):
        softwares = self.___load_softwares() if not evaluator else self.__load_evaluators()

        if identifier in softwares and not evaluator:
            return softwares[identifier]['TIRA_IMAGE_TO_EXECUTE'], softwares[identifier]['TIRA_COMMAND_TO_EXECUTE']
        if evaluator:
            for k, v in softwares.items():
                if k.startswith(identifier):
                    return v['TIRA_DATASET_ID'], v['TIRA_EVALUATION_IMAGE_TO_EXECUTE'], v['TIRA_EVALUATION_COMMAND_TO_EXECUTE']

        raise ValueError(f'There is no {("evaluator" if evaluator else "software")} identified by "{identifier}". Choices are: {sorted(list(softwares))}')

    def all_evaluated_appraoches(self) -> pd.DataFrame:
        id_to_software_name = {int(i['TIRA_SOFTWARE_ID'].split('docker-software-')[1]): i['TIRA_SOFTWARE_NAME'] for i in self.___load_softwares().values()}
        ret = []
        for evaluation in glob('*/*/*/evaluation'):
            job_dir = glob(evaluation + '/../job-executed-on*.txt')
            if len(job_dir) != 1:
                raise ValueError('Can not handle multiple job definitions: ', job_dir)

            job_definition = self.__load_job_data(job_dir[0])
            job_identifier = job_definition['TIRA_TASK_ID'] + '/' + job_definition['TIRA_VM_ID'] + '/' + id_to_software_name[int(job_definition['TIRA_SOFTWARE_ID'].split('docker-software-')[1])]

            for eval_run in glob(f"{evaluation}/*/output/"):
                try:
                    i = {'approach': job_identifier, 'dataset': job_definition['TIRA_DATASET_ID']}
                    i.update(self.__load_output(eval_run, evaluation=True))
                    ret += [i]
                except Exception:
                    pass

        return pd.DataFrame(ret)

    def docker_software(self, approach, software_id=None):
        ret = []

        if not os.path.exists(self.directory + '.tira/submitted-software.jsonl'):
            from tira.rest_api_client import Client as RestClient
            ret = RestClient().docker_software_details(approach)['context']
            return {'tira_image_name': ret['image'], 'command': ret['command'], 'id': None, 'ids_of_previous_stages': []}

        for _, i in self.all_softwares().iterrows():
            if (approach and i['approach'] == approach) or (software_id is not None and str(int(software_id)) == i['id']):
                ret += [{'tira_image_name': i['image'], 'command': i['command'], 'id': i['id'], 'ids_of_previous_stages': i['ids_of_previous_stages']}]

        if len(ret) == 1:
            return ret[0]

        raise ValueError(f'Could not find a unique software with approach="{approach}" or software_id="{software_id}". Found {ret}')

    def run_was_already_executed_on_dataset(self, approach, dataset):
        return self.get_run_execution_or_none(approach, dataset) is not None

    def get_run_output(self, approach, dataset, allow_without_evaluation=False):
        return self.directory + '/' + self.get_run_execution_or_none(approach, dataset)['run_id']

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None):
        task, team, software = approach.split('/')
        executions = [json.loads(i) for i in open(self.directory + '.tira/executions.jsonl')]

        executions = [i for i in executions if i['TIRA_TASK_ID'] == task and i['TIRA_VM_ID'] == team and i['TIRA_SOFTWARE_NAME'] == software and i['dataset'] == dataset]

        return None if len(executions) < 1 else {'task': task, 'dataset': dataset, 'team': team, 'run_id': executions[0]['output_directory']}

    def download_run(self, task, dataset, software, team=None, previous_stage=None, return_metadata=False):
        ret = self.get_run_execution_or_none(f'{task}/{team}/{software}', dataset, previous_stage)
        if not ret:
            raise ValueError(f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}", software="{software}", team={team}, previous_stage={previous_stage}')

        ret = self.directory + '/' + ret['run_id']

        if not os.path.exists(ret + '/run.txt'):
            ret = pd.read_json(ret + '/rerank.jsonl.gz', lines=True, orient='columns')
            ret['qid'] = ret['qid'].astype(str)
            ret['docno'] = ret['docno'].astype(str)
            ret['docid'] = ret['docno']
            ret['query'] = ret['qid']
        else:
            ret = pd.read_csv(ret + '/run.txt', sep='\\s+', names=["query", "q0", "docid", "rank", "score", "system"])
            ret['query'] = ret['query'].astype(str)
            ret['docid'] = ret['docid'].astype(str)
            ret['docno'] = ret['docid']
            ret['qid'] = ret['query']

        if return_metadata:
            return ret, 'run_id'
        else:
            return ret



from tira.local_client import Client

import pandas as pd
import unittest

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions

def keyphrase_extraction(docs):
    queries = pd.DataFrame([{'docno': str(i)} for i in docs])
    tira = Client('tests/resources/')
    query_segmentation =  tira.pt.transform_documents('ir-benchmarks/webis-keyphrase-extraction/BCExtractorFO', dataset='d1')
    
    ret = query_segmentation(queries)
    return {i['docno']: i['keyphrases'] for _, i in ret.iterrows()}


def doc_processor(docs, name):
    queries = pd.DataFrame([{'docno': str(i)} for i in docs])
    tira = Client('tests/resources/')
    query_segmentation =  tira.pt.transform_documents('ir-benchmarks/tira-ir-starters/' + name, dataset='d1')
    
    ret = query_segmentation(queries)
    return {i['docno']: i['key'] for _, i in ret.iterrows()}

class TestPtDocumentProcessingLoaderTest(unittest.TestCase):
    def test_for_loading_keyphrase_extraction_for_multiple_documents(self):
        expected = {
            'FT921-3964': ["increased economic aid","king hussein","increase aid","president george bush","tariq aziz iraq's deputy prime minister"],
            'LA061489-0137': ["public appearances gorbachev","crowds","government ministries gorbachev","place gorbachev","west german well-wishers"]}
        
        actual = keyphrase_extraction(['FT921-3964', 'LA061489-0137'])
        
        assert len(actual) == 2
        assert actual['FT921-3964'] == expected['FT921-3964']
        assert actual['LA061489-0137'] == expected['LA061489-0137']


    def test_for_loading_keyphrase_extraction_for_single_documents(self):
        expected = {"LA120390-0047": ["even friendly technology takeovers","largest technology takeover","computer industry takeover","statement sunday","attempts"]}
        
        actual = keyphrase_extraction(["LA120390-0047"])
        
        assert len(actual) == 1
        assert actual["LA120390-0047"] == expected["LA120390-0047"]


    def test_pyterrier_can_be_loaded(self):
        from tira.third_party_integrations import ensure_pyterrier_is_loaded
        ensure_pyterrier_is_loaded()

    def test_document_transformation_with_docno(self):
        expected = {'doc-01': 'value'}
        actual = doc_processor(['doc-01'], 'tiny-example-01')

        self.assertEqual(expected, actual)

    def test_document_transformation_with_doc_id(self):
        expected = {'doc-01': 'value'}
        actual = doc_processor(['doc-01'], 'tiny-example-02')

        self.assertEqual(expected, actual)



from tira.local_execution_integration import LocalExecutionIntegration
import unittest

# .. todo:: there are no assertions in this file

class TestExporSubmissionFromJupyterNotebook(unittest.TestCase):
    def test_non_existing_notebook_1(self):
        notebook = 'does-not-exist'

        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert actual is None


    def test_non_existing_notebook_2(self):
        notebook = 'does-not-exist/1/does-not-exist.ipynb'

        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert actual is None


    def test_notebook_submission_without_previous_stages(self):
        notebook = 'tests/resources/pyterrier-notebook-without-previous-stages.ipynb'

        expected = 'TIRA_COMMAND=/workspace/run-pyterrier-notebook.py --input ${TIRA_INPUT_DIRECTORY} --output ${TIRA_OUTPUT_DIRECTORY} --notebook /workspace/pyterrier-notebook-without-previous-stages.ipynb'
        actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

        assert expected == actual



from tira.local_execution_integration import LocalExecutionIntegration as Client
import tempfile
import unittest

# .. todo:: there are no assertions in this file

class TestDockerIntegration(unittest.TestCase):
    def test_export_of_file_from_bash_image(self):
        tira = Client()
        local_file = tempfile.NamedTemporaryFile().name
        tira.export_file_from_software('/etc/issue', local_file, image='bash:alpine3.16')
        
        actual = open(local_file).read()
        expected = '''Welcome to Alpine Linux 3.16
Kernel \\r on an \\m (\\l)

'''
        assert actual == expected

    def test_export_of_file(self):
        tira = Client()
        local_file = tempfile.NamedTemporaryFile().name
        tira.export_file_from_software('/etc/alpine-release', local_file, image='bash:alpine3.16')
        
        actual = open(local_file).read()
        expected = '''3.16.5\n'''
        assert actual == expected

    def test_extraction_of_entrypoint(self):
        expected = '/docker-entrypoint.sh'

        tira = Client()
        actual = tira.extract_entrypoint(image='bash:alpine3.16')

        assert actual == expected

    #def test_execution_of_software(self):
    #    tira = Client()
    #    tira.run(image='bash:alpine3.16', command='sleep 2s;', input_dir="/tmp/input", output_dir="/tmp/output")



from tira.third_party_integrations import register_rerank_data_to_ir_datasets, load_ir_datasets, ensure_pyterrier_is_loaded
import ir_datasets
import os
import unittest


# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions

class TestIRDatasets(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        register_rerank_data_to_ir_datasets('tests/sample-tirex-rerank-data', 'tirex-sample-dataset', 'vaswani')

    def test_for_existing_document_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load('tirex-sample-dataset')

        expected_text = "microwave spectroscopy  includes chapters on spectroscope technique\nand design on measurements on gases liquids and solids on nuclear\nproperties on molecular structure and on further possible applications\nof microwaves\n"

        assert expected_text == dataset.docs_store().get('8172').text
        assert '8172' == dataset.docs_store().get('8172').doc_id


    def test_for_existing_document_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load('tirex-sample-dataset')

        expected_text = "microwave spectroscopy  includes chapters on spectroscope technique\nand design on measurements on gases liquids and solids on nuclear\nproperties on molecular structure and on further possible applications\nof microwaves\n"

        assert expected_text == dataset.docs_store().get('8172').text
        assert '8172' == dataset.docs_store().get('8172').doc_id


    def test_for_non_existing_document_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load('tirex-sample-dataset')

        assert dataset.docs_store().get('does-not-exist') is None


    def test_for_non_existing_document_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load('tirex-sample-dataset')

        assert dataset.docs_store().get('does-not-exist') is None


    def test_for_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load('tirex-sample-dataset')

        assert len(list(dataset.queries_iter())) == 1
        assert 'MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES\n' == list(dataset.queries_iter())[0].text


    def test_for_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load('tirex-sample-dataset')

        assert len(list(dataset.queries_iter())) == 1
        assert 'MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES\n' == list(dataset.queries_iter())[0].text


    def test_for_scoreddocs_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        dataset = ir_datasets.load('tirex-sample-dataset')
        print(list(dataset.scoreddocs))
        assert len(list(dataset.scoreddocs)) == 10
        assert '8172' == list(dataset.scoreddocs)[0].doc_id
        assert '1' == list(dataset.scoreddocs)[0].query_id


    def test_for_scoreddocs_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load('tirex-sample-dataset')
        print(list(dataset.scoreddocs))
        assert len(list(dataset.scoreddocs)) == 10
        assert '8172' == list(dataset.scoreddocs)[0].doc_id
        assert '1' == list(dataset.scoreddocs)[0].query_id


    def test_for_scoreddocs_within_tira(self):
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/re-ranking-outputs'
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        dataset = ir_datasets.load('msmarco-passage/eval/small')
        print(list(dataset.scoreddocs))
        del os.environ['TIRA_INPUT_DATASET']
        assert len(list(dataset.scoreddocs)) == 3
        assert 'doc-1' == list(dataset.scoreddocs)[0].doc_id
        assert '1' == list(dataset.scoreddocs)[0].query_id
        assert 10 == list(dataset.scoreddocs)[0].score
        register_rerank_data_to_ir_datasets('tests/sample-tirex-rerank-data', 'tirex-sample-dataset', None)


    def test_loading_raw_ir_datasets_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load('cranfield')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 225
        assert queries['269'] == 'has a criterion been established for determining the axial compressor\nchoking line .'


    def test_loading_raw_ir_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load('cranfield')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 225
        assert queries['269'] == 'has a criterion been established for determining the axial compressor\nchoking line .'


    def test_loading_queries_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries['1'] == 'fox jumps above animal'


    def test_loading_queries_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_loading_queries_from_ir_datasets_from_custom_directory_2_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank-without-documents'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_loading_queries_from_ir_datasets_from_custom_directory_2_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank-without-documents'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        queries = {i.query_id: i.text for i in dataset.queries_iter()}

        assert len(list(dataset.queries_iter())) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_loading_documents_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        docs = {i.doc_id: i.text for i in dataset.docs_iter()}

        assert len(list(dataset.docs_iter())) == 5
        assert docs['pangram-02'] == 'Quick fox jumps nightly above wizard.'

    def test_loading_documents_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')
        docs = {i.doc_id: i.text for i in dataset.docs_iter()}

        assert len(list(dataset.docs_iter())) == 5
        assert docs['pangram-02'] == 'Quick fox jumps nightly above wizard.'

    def test_loading_docs_store_from_ir_datasets_from_custom_directory_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')

        assert dataset.docs_store()['pangram-03'].text == 'The jay, pig, fox, zebra and my wolves quack!'

    def test_loading_docs_store_from_ir_datasets_from_custom_directory_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        ir_datasets = load_ir_datasets()
        del os.environ['TIRA_INPUT_DATASET']
        dataset = ir_datasets.load('does-not-exist-and-is-not-used')

        assert dataset.docs_store()['pangram-03'].text == 'The jay, pig, fox, zebra and my wolves quack!'

    def test_no_patching_for_pyterrier_datasets_01(self):
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:cranfield')
        queries = {i['qid']: i['query'] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 225
        assert queries['269'] == 'has a criterion been established for determining the axial compressor choking line'

    def test_no_patching_for_pyterrier_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        import pyterrier as pt
        dataset = pt.get_dataset('irds:cranfield')
        queries = {i['qid']: i['query'] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 225
        assert queries['269'] == 'has a criterion been established for determining the axial compressor choking line'

    def test_patching_for_pyterrier_datasets_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics().iterrows()}

        assert len(list(dataset.get_topics('title'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_patching_for_pyterrier_datasets_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics().iterrows()}

        assert len(list(dataset.get_topics('title'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_patching_for_pyterrier_datasets_for_text_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics('text').iterrows()}

        assert len(list(dataset.get_topics('text'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_patching_for_pyterrier_datasets_for_text_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics('text').iterrows()}

        assert len(list(dataset.get_topics('text'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_patching_for_pyterrier_datasets_for_title_queries_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics('title').iterrows()}

        assert len(list(dataset.get_topics('title'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_patching_for_pyterrier_datasets_for_title_queries_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/sample-input-full-rank'
        import ir_datasets
        l = ir_datasets.load
        ensure_pyterrier_is_loaded()
        import pyterrier as pt
        dataset = pt.get_dataset('irds:does-not-exist-and-is-not-used')
        ir_datasets.load = l
        del os.environ['TIRA_INPUT_DATASET']

        queries = {i['qid']: i['query'] for _, i in dataset.get_topics('title').iterrows()}

        assert len(list(dataset.get_topics('title'))) == 2
        assert queries['1'] == 'fox jumps above animal'

    def test_loading_ir_dataset_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert dataset.has_docs(), "dataset has no documents"
        assert dataset.has_queries(), "dataset has no queries"
        assert dataset.has_qrels(), "dataset has no qrels"

    def test_loading_ir_dataset_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert dataset.has_docs(), "dataset has no documents"
        assert dataset.has_queries(), "dataset has no queries"
        assert dataset.has_qrels(), "dataset has no qrels"
        
    def test_loading_qrels_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert len([i for i in dataset.qrels_iter()]) == 2635
        assert '''TrecQrel(query_id='1', doc_id='2005.ipm_journal-ir0anthology0volumeA41A1.7', relevance=1, iteration='0')''' == str([i for i in dataset.qrels_iter()][0])

    def test_loading_qrels_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert len([i for i in dataset.qrels_iter()]) == 2635
        assert '''TrecQrel(query_id='1', doc_id='2005.ipm_journal-ir0anthology0volumeA41A1.7', relevance=1, iteration='0')''' == str([i for i in dataset.qrels_iter()][0])

    def test_loading_docs_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert len([i for i in dataset.docs_iter()]) == 53673
        docsstore = dataset.docs_store()

        assert docsstore.get('2005.ipm_journal-ir0anthology0volumeA41A1.7').text.startswith('A probabilistic model for stemmer generation AbstractIn this paper')

    def test_loading_docs_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert len([i for i in dataset.docs_iter()]) == 53673
        docsstore = dataset.docs_store()

        assert docsstore.get('2005.ipm_journal-ir0anthology0volumeA41A1.7').text.startswith('A probabilistic model for stemmer generation AbstractIn this paper')

    def test_loading_topics_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.topics_file('ir-lab-jena-leipzig-wise-2023/training-20231104-training')
        
        assert dataset.endswith('/training-20231104-training/truth-data/queries.xml')

    def test_loading_topics_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.topics_file('ir-lab-jena-leipzig-wise-2023/training-20231104-training')
        
        assert dataset.endswith('/training-20231104-training/truth-data/queries.xml')

    def test_loading_topics_via_rest_api_from_rerank_dataset_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        os.environ['TIRA_INPUT_DATASET'] = 'tests/resources/re-ranking-outputs'
        ir_datasets = load_ir_datasets()
        dataset = ir_datasets.load('workshop-on-open-web-search/re-ranking-20231027-training')
        del os.environ['TIRA_INPUT_DATASET']

        assert len([i for i in dataset.queries_iter()]) == 2
        assert len([i for i in dataset.docs_iter()]) == 3

    def test_loading_queries_via_rest_api_from_tira_01(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=False)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert 68 == len(list(dataset.queries_iter()))
        print(str(list(dataset.queries_iter())[0]))
        assert '''TirexQuery(query_id='1', text='retrieval system improving effectiveness', title='retrieval system improving effectiveness', query='retrieval system improving effectiveness', description='What papers focus on improving the effectiveness of a retrieval system?', narrative='Relevant papers include research on what makes a retrieval system effective and what improves the effectiveness of a retrieval system. Papers that focus on improving something else or improving the effectiveness of a system that is not a retrieval system are not relevant.')''' == str(list(dataset.queries_iter())[0])

    def test_loading_queries_via_rest_api_from_tira_02(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        from tira.third_party_integrations import ir_datasets
        dataset = ir_datasets.load('ir-lab-jena-leipzig-sose-2023/iranthology-20230618-training')
        
        assert 68 == len(list(dataset.queries_iter()))
        print(str(list(dataset.queries_iter())[0]))
        assert '''TirexQuery(query_id='1', text='retrieval system improving effectiveness', title='retrieval system improving effectiveness', query='retrieval system improving effectiveness', description='What papers focus on improving the effectiveness of a retrieval system?', narrative='Relevant papers include research on what makes a retrieval system effective and what improves the effectiveness of a retrieval system. Papers that focus on improving something else or improving the effectiveness of a system that is not a retrieval system are not relevant.')''' == str(list(dataset.queries_iter())[0])

    def test_patching_for_pyterrier_datasets_to_tira(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt
        dataset = pt.get_dataset('irds:workshop-on-open-web-search/retrieval-20231027-training')
        queries = {i['qid']: i['title'] for _, i in dataset.get_topics().iterrows()}

        assert len(dataset.get_topics()) == 3
        assert queries['1'] == 'hubble telescope achievements'

    def test_patching_for_pyterrier_datasets_with_qrels_to_tira(self):
        ensure_pyterrier_is_loaded(patch_ir_datasets=True)
        import pyterrier as pt
        dataset = pt.get_dataset('irds:workshop-on-open-web-search/retrieval-20231027-training')
        qrels =[i.to_dict() for _, i in dataset.get_qrels().iterrows()]

        assert len(dataset.get_qrels()) == 3
        assert qrels[0] == {'docno': 'doc-1', 'iteration': '0', 'label': 1, 'qid': '1'}



from tira.local_client import Client

import pandas as pd
import unittest

# .. todo:: this file still uses "unclean" assertions and should be converted to use unittest assertions

def retrieval_submission(queries):
    queries = pd.DataFrame([{'qid': str(i)} for i in queries])
    tira = Client('tests/resources/')
    retriever =  tira.pt.from_retriever_submission('ir-benchmarks/tira-ir-starters/retriever-for-re-ranking', dataset='d1')
    
    ret = retriever(queries)
    return [{'qid': i['qid'], 'docno': i['docno'], 'score': i['score'], 'text': i['text']} for _, i in ret.iterrows()]

class TestPtFromRetrieverForReRankingSubmissionLoader(unittest.TestCase):
    def test_for_multiple_queries(self):
        expected = [
            {'qid': '1', 'docno': 'doc-1', 'score':10, 'text': "Text of doc-1"},
            {'qid': '1', 'docno': 'doc-2', 'score':9, 'text': "Text of doc-2"},
            {'qid': '3', 'docno': 'doc-3', 'score': 1, 'text': "Text of doc-3"},
        ]
        
        actual = retrieval_submission([1, 3])

        assert len(actual) == 3
        assert actual[0] == expected[0]
        assert actual[1] == expected[1]
        assert actual[2] == expected[2]

    def test_for_single_query_no_1(self):
        expected = [
            {'qid': '1', 'docno': 'doc-1', 'score':10, 'text': "Text of doc-1"},
            {'qid': '1', 'docno': 'doc-2', 'score':9, 'text': "Text of doc-2"},
        ]
        
        actual = retrieval_submission([1])

        assert len(actual) == 2
        assert actual[0] == expected[0]
        assert actual[1] == expected[1]

    def test_for_single_query_no_3(self):
        expected = [
            {'qid': '3', 'docno': 'doc-3', 'score': 1, 'text': "Text of doc-3"},
        ]
        
        actual = retrieval_submission([3])

        assert len(actual) == 1
        assert actual[0] == expected[0]



from copy import deepcopy
from typing import NamedTuple
from tira.io_utils import all_lines_to_pandas, stream_all_lines
from tira.tirex import IRDS_TO_TIREX_DATASET, TIREX_DATASETS
import os
import logging

original_ir_datasets_load = None
try:
    import ir_datasets as original_ir_datasets
    original_ir_datasets_load = original_ir_datasets.load
except ImportError:
    pass

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .tira_client import TiraClient


class TirexQuery(NamedTuple):
    query_id: str
    text: str
    title: str
    query: str
    description: str
    narrative: str

    def default_text(self):
        """
        title
        """
        return self.title


def register_dataset_from_re_rank_file(ir_dataset_id, df_re_rank, original_ir_datasets_id=None):
    """
    Load a dynamic ir_datasets integration from a given re_rank_file.
    The dataset will be registered for the id ir_dataset_id.
    The original_ir_datasets_id is used to infer the class of documents, qrels, and queries.
    """
    import ir_datasets
    from ir_datasets.datasets.base import Dataset
    original_dataset = ir_datasets.load(original_ir_datasets_id) if original_ir_datasets_id else None

    docs = __docs(df_re_rank, original_dataset, True)
    queries = __queries(df_re_rank, original_dataset)
    qrels = __lazy_qrels(None, original_dataset)
    scoreddocs = __scored_docs(df_re_rank, original_dataset)

    dataset = Dataset(docs, queries, qrels, scoreddocs)
    dataset.metadata = None
    ir_datasets.registry._allow_overwrite = True
    ir_datasets.registry.register(ir_dataset_id, dataset)

    __check_registration_was_successful(ir_dataset_id, original_ir_datasets_id is None)


def translate_irds_id_to_tirex(dataset):
    if type(dataset) is not str:
        if hasattr(dataset, 'irds_ref'):
            return translate_irds_id_to_tirex(dataset.irds_ref().dataset_id())
        else:
            raise ValueError(f'I can not handle {dataset}.')
    return IRDS_TO_TIREX_DATASET.get(dataset, dataset)


def __docs(input_file, original_dataset, load_default_text):
    from ir_datasets.formats import BaseDocs, GenericDoc

    class DynamicDocs(BaseDocs):
        def __init__(self, input_file, load_default_text):
            self.docs = None
            self.input_file = input_file
            self.load_default_text = load_default_text

        def docs_iter(self):
            return self.stream_docs()

        def docs_count(self):
            return sum(1 for _ in self.stream_docs())

        def docs_store(self):
            return self.__parsed_docs()

        def __parsed_docs(self):
            if self.docs is None:
                d = {}
                for i in self.stream_docs():
                    d[i.doc_id] = i

                self.docs = d

            return self.docs

        def stream_docs(self):
            already_covered = set()
            for i in stream_all_lines(self.get_input_file(), self.load_default_text):
                if i['docno'] not in already_covered:
                    already_covered.add(i['docno'])
                    yield GenericDoc(doc_id=i['docno'], text=i['text'])

        def get_input_file(self):
            if type(self.input_file) is str:
                return self.input_file

            ret = self.input_file()
            if os.path.isfile(ret + '/documents.jsonl.gz'):
                return ret + '/documents.jsonl.gz'
            else:
                return ret + '/documents.jsonl'

    return DynamicDocs(input_file, load_default_text)


def __lazy_qrels(input_file, original_qrels):
    from ir_datasets.formats import BaseQrels, TrecQrel, TrecQrels
    from ir_datasets.util.download import LocalDownload

    if input_file is None and original_qrels is None:
        return

    class QrelsFromTira(BaseQrels):
        def __init__(self, input_file, original_qrels):
            self.qrels = None if not original_qrels else original_qrels
            self.input_file = input_file
            self.original_qrels = original_qrels

        def qrels_cls(self):
            return TrecQrel if not self.original_qrels else self.original_qrels.qrels_cls()

        def qrels_iter(self):
            if not self.qrels:
                qrels_file = self.input_file() + '/qrels.txt'

                self.qrels = TrecQrels(LocalDownload(qrels_file), {})

            return self.qrels.qrels_iter()

        def qrels_defs(self):
            self.qrels_iter()
            return self.qrels.qrels_defs()

    return QrelsFromTira(input_file, original_qrels)


def __queries(input_file, original_dataset):
    from ir_datasets.formats import BaseQueries

    class DynamicQueries(BaseQueries):
        def __init__(self, input_file):
            self.queries = None
            self.input_file = input_file

        def queries_iter(self):
            return deepcopy(self.__get_queries()).values().__iter__()

        def queries_cls(self):
            return TirexQuery

        def get_input_file(self):
            if type(self.input_file) is str:
                return self.input_file

            return self.input_file() + '/queries.jsonl'

        def __get_queries(self):
            if self.queries is None:
                ret = {}
                for i in stream_all_lines(self.get_input_file(), False):
                    orig_query = None if 'original_query' not in i else i['original_query']
                    description = None if (not orig_query or type(orig_query) is not dict or 'description' not in orig_query) else orig_query['description']
                    narrative = None if (not orig_query or type(orig_query) is not dict or 'narrative' not in orig_query) else orig_query['narrative']
                    if i['qid'] not in ret:
                        ret[i['qid']] = TirexQuery(query_id=i['qid'],  text=i['query'], query=i['query'], title=i['query'], description=description, narrative=narrative)

                self.queries = ret

            return self.queries

    return DynamicQueries(input_file)


def __scored_docs(input_file, original_dataset):
    from ir_datasets.formats import BaseScoredDocs

    class GenericScoredDocWithRank(NamedTuple):
        query_id: str
        doc_id: str
        score: float
        rank: int

    class DynamicScoredDocs(BaseScoredDocs):
        def __init__(self, docs):
            self.docs = docs

        def scoreddocs_iter(self):
            return deepcopy(self.docs).__iter__()

        def scoreddocs_count(self) -> int:
            return len(docs)

    docs = []

    for i in stream_all_lines(input_file, True):
        docs += [GenericScoredDocWithRank(i['qid'], i['docno'], i['score'], i['rank'])]

    return DynamicScoredDocs(docs)


def static_ir_dataset(directory, existing_ir_dataset=None):
    from ir_datasets.datasets.base import Dataset
    if existing_ir_dataset is None:
        queries_file = directory + '/queries.jsonl'
        docs_file = directory
        if os.path.isfile(docs_file + '/documents.jsonl.gz'):
            docs_file = docs_file + '/documents.jsonl.gz'
        else:
            docs_file = docs_file + '/documents.jsonl'

        docs = __docs(docs_file, None, True)
        queries = __queries(queries_file, None)
        return static_ir_dataset(directory, Dataset(docs, queries))

    class StaticIrDatasets():
        def load(self, dataset_id):
            print(f'Load ir_dataset from "{directory}" instead of "{dataset_id}" because code is executed in TIRA.')
            return existing_ir_dataset

        def topics_file(self, dataset_id):
            return directory + '/queries.xml'

    return StaticIrDatasets()


def ir_dataset_from_tira_fallback_to_original_ir_datasets():
    from ir_datasets.datasets.base import Dataset

    def get_download_dir_from_tira(tira_path, truth_dataset) -> "TiraClient":
        if len(tira_path.split('/')) != 2:
            logging.info(f'Please pass tira_path as <task>/<tira-dataset>. Got {tira_path}')
            raise ValueError(f'Please pass tira_path as <task>/<tira-dataset>. Got {tira_path}')

        from tira.rest_api_client import Client as RestClient
        task, dataset = tira_path.split('/')
        return RestClient().download_dataset(task, dataset, truth_dataset=truth_dataset)

    class IrDatasetsFromTira():
        def load(self, dataset_id):
            try:
                return original_ir_datasets_load(dataset_id)
            except Exception:
                logging.info(f'Load ir_dataset "{dataset_id}" from tira.')
                docs = self.lazy_docs(lambda: get_download_dir_from_tira(dataset_id, False), None, True)
                queries = self.lazy_queries(lambda: get_download_dir_from_tira(dataset_id, True), None)
                qrels = self.lazy_qrels(lambda: get_download_dir_from_tira(dataset_id, True), None)
                return Dataset(docs, queries, qrels)

        def topics_file(self, tira_path):
            return get_download_dir_from_tira(tira_path, True) + '/queries.xml'

    ret = IrDatasetsFromTira()
    ret.lazy_docs = __docs
    ret.lazy_queries = __queries
    ret.lazy_qrels = __lazy_qrels

    return ret


def __check_registration_was_successful(ir_dataset_id, ignore_qrels=True):
    import ir_datasets
    dataset = ir_datasets.load(ir_dataset_id)

    assert dataset.has_docs(), "dataset has no documents"
    assert dataset.has_queries(), "dataset has no queries"
    assert ignore_qrels or dataset.has_qrels(), "dataset has no qrels"
    assert dataset.has_scoreddocs(), "dataset has no scored_docs"
